def default_retry_handler(
    e: Exception, base_seconds_delay: int = 0, *args, **kwargs
) -> int:
    """Package default retry handler

    Will return return nr of seconds to sleep

    Args:
        e(Exception): the exception that occured
        base_seconds_delay(int): nr of seconds to sleep for between retries. Defaults to 0
    """

    return base_seconds_delay



from yet_another_retry.retry_handlers import exponential_backoff


def test_exponential_backoff():

    try:

        sleep_seconds = [
            exponential_backoff(e=Exception, attempt=i) for i in range(1, 10, 1)
        ]

        # with default settings of exponential_backoff this should be the result
        assert sleep_seconds == [2, 4, 8, 16, 32, 64, 128, 256, 512]
    except:
        assert False



from yet_another_retry import retry


@retry()
def my_function():
    """This function will use the default retry and exception handlers and fail the default nr of times"""

    print("Raising an error")
    raise Exception("Raising an error")


my_function()



from yet_another_retry import retry


@retry(tries=10, base_seconds_delay=5)
def my_function():
    """This function will delay for 5 seconds and retry 10 as per config"""

    print("Raising an error")
    raise Exception("Raising an error")


my_function()



from yet_another_retry.retry_handlers.default_retry_handler import default_retry_handler
from yet_another_retry.retry_handlers.sleep_attempt_seconds import sleep_attempt_seconds
from yet_another_retry.retry_handlers.exponential_backoff import exponential_backoff

__all__ = ["default_retry_handler", "sleep_attempt_seconds", "exponential_backoff"]



import pytest
from yet_another_retry.exception_handlers import default_exception_handler


def test_default_retry_handler():
    try:
        default_exception_handler(e=Exception)
        assert False
    except Exception:
        assert True



from yet_another_retry import retry


@retry()
def my_function(retry_config: dict):

    attempt = retry_config["attempt"]
    print(f"This is attempt number: {attempt}")
    raise Exception("This is an exception")


my_function()



def sleep_attempt_seconds(
    e: Exception, attempt: int, base_seconds_delay: int = 0, *args, **kwargs
) -> int:
    """Retry handler that returns the attempt number as delay

    Will return return nr of seconds to sleep as attempt + base_seconds_delay

    Args:
        e(Exception): the exception that occured
        attempt(int): Attempt number, is supplied by the decorator.
        base_seconds_delay(int|float): Number of seconds to sleep on top of the attempt. Defaults to 0
    """

    sleep_time = attempt + base_seconds_delay

    return sleep_time



from yet_another_retry import retry
from yet_another_retry.retry_handlers import exponential_backoff


@retry(
    retry_handler=exponential_backoff,
    tries=5,
    base_seconds_delay=1,
    exponential_factor=3,
    max_delay_seconds=1800,
    jitter_range=10,
)
def my_function(retry_config: dict):
    print(f"This is attempt number: {retry_config['attempt']}")
    print(f"Last sleep was {retry_config['previous_delay']}")
    raise Exception("This is an exception")


my_function()



from typing import Callable
import time
import inspect
from yet_another_retry.retry_handlers import default_retry_handler
from yet_another_retry.exception_handlers import default_exception_handler
import functools


def retry(
    retry_exceptions: Exception | tuple[Exception] = Exception,
    fail_on_exceptions: Exception | tuple[Exception] = (),
    tries: int = 3,
    base_seconds_delay: float | int = 0,
    retry_handler: Callable = default_retry_handler,
    exception_handler: Callable = default_exception_handler,
    raise_final_exception: bool = True,
    *args,
    **kwargs,
) -> Callable:
    """Decorator for retrying a function

    Args:
        retry_exceptions(tuple[Exception]): An Exception or tuple of exceptions to retry. All other exceptions will fail. Defaults to Exception, meaning all exceptions are retried unless this value is modified.
        fail_on_exceptions(tuple[Exception]): An Exception or tuple of exception to not retry but instead raise error if it occures. Defaults to ()
        tries(int): Maximum number of retries to attempt. Defaults to 3
        base_seconds_delay(int | float): number of seconds to delay, used by default retry handler and other built in handlers. Defaults to 0
        retry_handler(Callable): Callable function to run in case of retries. Defaults to default retry_handler function
        exception_handler(Callable): Callable function to run in case of erroring out, either by reaching max tries +1 or hitting a fail_on_exception exception. Defaults to default exception_handler function.
        raise_error(bool): If set to false the decorator itself will not raise the error but expect the handler to do it. Default is True
        **kwargs: any extra kwargs gets added as input to handlers and will also be added to the retry_config
    """

    def decorator(func: Callable) -> Callable:
        # get the signature of the decorated function
        sig = inspect.signature(func)
        # check if "retry_config" is in the signature so we know later to send the retry_config or not
        add_retry_config = True if "retry_config" in sig.parameters else False

        def wrapper(*func_args, **func_kwargs):
            retry_config = {
                "retry_exceptions": retry_exceptions,
                "fail_on_exceptions": fail_on_exceptions,
                "tries": tries,
                "base_seconds_delay": base_seconds_delay,
                "retry_handler": retry_handler,
                "exception_handler": exception_handler,
                "raise_final_exception": raise_final_exception,
                "attempt": 0,
                "previous_delay": 0,
                **kwargs,
            }

            if add_retry_config:
                func_kwargs["retry_config"] = retry_config
            # makes a copy of the input parameters to the decorated function so we can pass it to the function
            # without accidentally sending any extra args/kwargs from the decorator
            # bound_args = sig.bind_partial(*args, **kwargs)
            # bound_args.apply_defaults()

            # placeholder for previous delay on each loop
            previous_delay = 0

            for i in range(1, tries + 1):
                try:
                    # add values to be passed to the handlers from the current loop
                    kwargs["attempt"] = i
                    kwargs["previous_delay"] = previous_delay

                    if add_retry_config:
                        func_kwargs["retry_config"]["attempt"] = i

                    return func(*func_args, **func_kwargs)

                except fail_on_exceptions as e:

                    if exception_handler:
                        exception_handler(e, *args, **kwargs)
                    if raise_final_exception:
                        raise e

                except retry_exceptions as e:
                    if i == tries:
                        if exception_handler:
                            exception_handler(e, *args, **kwargs)
                        if raise_final_exception:
                            raise e
                    delay_seconds = retry_handler(e, *args, **kwargs)

                    if add_retry_config:
                        retry_config["previous_delay"] = delay_seconds

                    # the return from the retry handler must be an int or a float
                    if not isinstance(delay_seconds, (int, float)):
                        raise TypeError(
                            f"The retry_handler did not return an int or float. Can not use {type(delay_seconds)} as input to sleep"
                        )
                    time.sleep(delay_seconds)

        return wrapper

    return decorator



"""Example of a custom exception handler.
Exception handler must have the following parameters:

e: Exception
retry_config: dict
-- any custom input
*args
**kwargs

The decorator might be submitting things that your custom handler is not expecting so *args and **kwargs are required
"""

from yet_another_retry import retry


def custom_exception_handler(
    e: Exception,
    attempt: int,
    special_message_on_fail: str = None,
    **kwargs,
) -> None:
    """Custom handler that accepts the config and any other extra parameters required

    Args:
        e(Exception): the exception raised
        attempt(int): the attempt number
        special_message_on_fail(str): an extra message to print before raising the error

    Raises:
        Exception: the exception provided
    """

    print(f"Failed to do something on attempt nr {attempt}")
    print(special_message_on_fail)
    print(f"The error was a {e.__class__.__name__} Exception")
    raise e


@retry(
    exception_handler=custom_exception_handler,
    special_message_on_fail="Message to print on fail",
)
def my_function():
    raise Exception("This is an Exception")


my_function()



import argparse
from . import __version__
import sys

def main():
    parser = argparse.ArgumentParser(description="Yet Another Retry")
    parser.add_argument("--version", action="version", version=f"%(prog)s {__version__}", help="Prints current installed version")


    if len(sys.argv) == 1:
        parser.print_help()

    args = parser.parse_args()
    

if __name__ == "__main__":
    main()


import pytest
from yet_another_retry.retry_handlers import default_retry_handler


def test_default_retry_handler():

    try:
        delay_seconds = default_retry_handler(e=Exception)

        assert delay_seconds == 0
    except:
        assert False



from yet_another_retry import retry


def test_retry():

    try:
        assert function_to_retry()
    except:
        assert False


@retry()
def function_to_retry(retry_config: dict):

    attempt = retry_config["attempt"]
    tries = retry_config["tries"]
    if attempt == tries:
        return True

    raise Exception("This is an exception")



def default_exception_handler(e: Exception, *args, **kwargs):
    """Base function for handling exception
    Args:
        e(Exception): the exception to raise

    Raises:
        Exception: the exception supplied.
    """

    raise e



from yet_another_retry.retry_handlers import sleep_attempt_seconds
import random


def test_sleep_attempt_seconds():

    try:
        random_attempt = random.randint(1, 10)
        sleep_seconds = sleep_attempt_seconds(e=Exception, attempt=random_attempt)

        assert sleep_seconds == random_attempt
    except:
        assert False



import random


def exponential_backoff(
    e: Exception,
    attempt: int,
    base_seconds_delay: float | int = 1,
    exponential_factor: float | int = 2,
    max_delay_seconds: float | int = None,
    jitter_range: float | int = None,
    *args,
    **kwargs,
) -> int:
    """Retry handler increases the sleep time exponentially.

    How long to sleep is calculcated as::

        (base_delay x exponential_factor^attempt) + jitter

    Default parameters will result in a standard base-2 exponential increase::

        2, 4, 8, 16, 32, 64, 128....

    Args:
        e(Exception): The exception that occured. Defaults to Exception.
        attempt(int): Attempt number, is passed from the decorator on each retry.
        base_delay_seconds(float | int): Base of the sleep calculation. Defaults to 1.
        exponential_factor(float | int): Multiplyer for the sleep calculation. Defaults to 2
        max_delay_seconds(float | int, optional): The max seconds to sleep between tries. If None no upper limit. Defaults to None.
        jitter_range(float | int, optional): If supplied, a random delay will be added between jitter_range and negative jitter_range (-jitter_range) with a step of 0.1 If None no jitter is added. Defaults to None

    Returns:
        float: Number of seconds to sleep

    """

    sleep_delay = base_seconds_delay * (exponential_factor**attempt)

    if max_delay_seconds and sleep_delay > max_delay_seconds:
        sleep_delay = max_delay_seconds

    if jitter_range and jitter_range > 0:
        # float between -jitter_max_seconds and jitter_max_seconds as a float added to the delay'
        # there are other functions to do this but to produce floats they are either slower or very convoluted.
        jitter = random.randint(-jitter_range * 10, jitter_range * 10) / 10
        sleep_delay += jitter

    # if we for some reason went below 0 we have to return 0
    if sleep_delay < 0:
        return 0

    return round(sleep_delay, 2)



"""Example of a custom retry handler.
Retry handler must have the following parameters:

e: Exception
-- any custom input
*args
**kwargs

The decorator might be submitting things that your custom handler is not expecting so **kwargs is a required input parameter

All retry handlers must return an integer which is the delay/sleep time in seconds.
"""

from yet_another_retry import retry


def custom_retry_handler(
    e: Exception, attempt: int, sleep_modifier: int = 1, **kwargs
) -> int:
    """Custom handler that accepts the config and any other extra parameters required

    Args:
        e(Exception): the exception raised
        sleep_modifier(int): a modifier for the sleep delay

    Returns:
        int: the time to sleep in seconds
    """

    print(f"This is attempt nr {attempt}")
    print(f"The error was a {e.__class__.__name__} Exception")
    delay = attempt * sleep_modifier
    print(f"Will sleep for {delay} seconds")

    return delay


@retry(retry_handler=custom_retry_handler, sleep_modifier=5)
def my_function():
    raise Exception("This is an exception")


my_function()
